@@ -36,6 +36,12 @@ public class DebuggerTests
 	public static string runtime = Environment.GetEnvironmentVariable ("DBG_RUNTIME");
 	public static string agent_args = Environment.GetEnvironmentVariable ("DBG_AGENT_ARGS");
 
+	Event GetNextEvent () {
+		var es = vm.GetNextEventSet ();
+		Assert.AreEqual (1, es.Events.Length);
+		return es [0];
+	}
+
 	void Start (string[] args) {
 		if (!listening) {
 			var pi = new Diag.ProcessStartInfo ();
@@ -51,9 +57,10 @@ public class DebuggerTests
 			vm = VirtualMachineManager.Listen (new IPEndPoint (IPAddress.Any, 10000));
 		}
 
-		vm.EnableEvents (EventType.AssemblyLoad);
+		var load_req = vm.CreateAssemblyLoadRequest ();
+		load_req.Enable ();
 
-		Event vmstart = vm.GetNextEvent ();
+		Event vmstart = GetNextEvent ();
 		Assert.AreEqual (EventType.VMStart, vmstart.EventType);
 
 		vm.Resume ();
@@ -65,7 +72,7 @@ public class DebuggerTests
 
 		/* Find out the entry point */
 		while (true) {
-			e = vm.GetNextEvent ();
+			e = GetNextEvent ();
 
 			if (e is AssemblyLoadEvent) {
 				AssemblyLoadEvent ae = (AssemblyLoadEvent)e;
@@ -76,6 +83,8 @@ public class DebuggerTests
 
 			vm.Resume ();
 		}
+
+		load_req.Disable ();
 	}
 
 	BreakpointEvent run_until (string name) {
@@ -88,7 +97,7 @@ public class DebuggerTests
 
 		while (true) {
 			vm.Resume ();
-			e = vm.GetNextEvent ();
+			e = GetNextEvent ();
 			if (e is BreakpointEvent)
 				break;
 		}
@@ -104,7 +113,7 @@ public class DebuggerTests
 		req.Enable ();
 
 		vm.Resume ();
-		Event e = vm.GetNextEvent ();
+		Event e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 
 		req.Disable ();
@@ -152,7 +161,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 		while (true) {
-			Event e = vm.GetNextEvent ();
+			Event e = GetNextEvent ();
 
 			if (e is VMDeathEvent)
 				break;
@@ -172,7 +181,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.AreEqual (EventType.Breakpoint, e.EventType);
 		Assert.IsTrue (e is BreakpointEvent);
 		Assert.AreEqual (m.Name, (e as BreakpointEvent).Method.Name);
@@ -186,8 +195,6 @@ public class DebuggerTests
 
 	[Test]
 	public void BreakpointsSameLocation () {
-		Event e;
-
 		MethodMirror m = entry_point.DeclaringType.GetMethod ("bp2");
 		Assert.IsNotNull (m);
 
@@ -196,13 +203,13 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
-		Assert.IsTrue (e is BreakpointEvent);
-		Assert.AreEqual (m, (e as BreakpointEvent).Method);
+		var es = vm.GetNextEventSet ();
+		Assert.AreEqual (2, es.Events.Length);
+		Assert.IsTrue (es [0] is BreakpointEvent);
+		Assert.AreEqual (m, (es [0] as BreakpointEvent).Method);
 
-		e = vm.GetNextEvent ();
-		Assert.IsTrue (e is BreakpointEvent);
-		Assert.AreEqual (m.Name, (e as BreakpointEvent).Method.Name);
+		Assert.IsTrue (es [1] is BreakpointEvent);
+		Assert.AreEqual (m.Name, (es [1] as BreakpointEvent).Method.Name);
 	}
 
 	[Test]
@@ -221,21 +228,21 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.AreEqual (EventType.Breakpoint, e.EventType);
 		Assert.AreEqual (m.Name, (e as BreakpointEvent).Method.Name);
 
 		vm.Resume ();
 
 		/* Non-shared instance */
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.AreEqual (EventType.Breakpoint, e.EventType);
 		Assert.AreEqual (m2.Name, (e as BreakpointEvent).Method.Name);
 
 		vm.Resume ();
 
 		/* Shared instance */
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.AreEqual (EventType.Breakpoint, e.EventType);
 		Assert.AreEqual (m2.Name, (e as BreakpointEvent).Method.Name);
 	}
@@ -256,19 +263,19 @@ public class DebuggerTests
 		/* Run until bp4 */
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
-		Assert.AreEqual (EventType.Breakpoint, e.EventType);
-		Assert.AreEqual (m.Name, (e as BreakpointEvent).Method.Name);
-		e = vm.GetNextEvent ();
-		Assert.AreEqual (EventType.Breakpoint, e.EventType);
-		Assert.AreEqual (m.Name, (e as BreakpointEvent).Method.Name);
+		var es = vm.GetNextEventSet ();
+		Assert.AreEqual (2, es.Events.Length);
+		Assert.AreEqual (EventType.Breakpoint, es [0].EventType);
+		Assert.AreEqual (m.Name, (es [0] as BreakpointEvent).Method.Name);
+		Assert.AreEqual (EventType.Breakpoint, es [1].EventType);
+		Assert.AreEqual (m.Name, (es [1] as BreakpointEvent).Method.Name);
 
 		/* Clear one of them */
 		req1.Disable ();
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.AreEqual (EventType.Breakpoint, e.EventType);
 		Assert.AreEqual (m.Name, (e as BreakpointEvent).Method.Name);
 
@@ -277,7 +284,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.AreEqual (EventType.Breakpoint, e.EventType);
 		Assert.AreEqual (m2.Name, (e as BreakpointEvent).Method.Name);
 	}
@@ -298,7 +305,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (!(e is BreakpointEvent));
 		if (e is VMDeathEvent)
 			vm = null;
@@ -315,7 +322,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.AreEqual (EventType.Breakpoint, e.EventType);
 		Assert.IsTrue (e is BreakpointEvent);
 		Assert.AreEqual (m.Name, (e as BreakpointEvent).Method.Name);
@@ -332,13 +339,13 @@ public class DebuggerTests
 
 		// Step into ss1
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("ss1", (e as StepEvent).Method.Name);
 
 		// Step out of ss1
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("single_stepping", (e as StepEvent).Method.Name);
 
@@ -349,7 +356,7 @@ public class DebuggerTests
 
 		// Step over ss2
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("single_stepping", (e as StepEvent).Method.Name);
 
@@ -360,7 +367,7 @@ public class DebuggerTests
 
 		// Step into ss3
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("ss3", (e as StepEvent).Method.Name);
 
@@ -371,7 +378,7 @@ public class DebuggerTests
 
 		// Step back into single_stepping
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("single_stepping", (e as StepEvent).Method.Name);
 
@@ -382,7 +389,7 @@ public class DebuggerTests
 
 		// Step into ss3_2 ()
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("ss3_2", (e as StepEvent).Method.Name);
 
@@ -393,7 +400,7 @@ public class DebuggerTests
 
 		// Step over ss3_2_2 ()
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("ss3_2", (e as StepEvent).Method.Name);
 
@@ -403,7 +410,7 @@ public class DebuggerTests
 
 		// Step back into single_stepping () with the new request
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("single_stepping", (e as StepEvent).Method.Name);
 
@@ -414,7 +421,7 @@ public class DebuggerTests
 
 		// Step into ss4 ()
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("ss4", (e as StepEvent).Method.Name);
 
@@ -426,7 +433,7 @@ public class DebuggerTests
 
 		// Step over ss1 (); ss1 ();
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 
 		// Step into ss2 ()
@@ -435,7 +442,7 @@ public class DebuggerTests
 		req.Enable ();
 
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("ss2", (e as StepEvent).Method.Name);
 
@@ -450,7 +457,7 @@ public class DebuggerTests
 
 		// Step into is_even, skipping the linq stuff
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("is_even", (e as StepEvent).Method.Name);
 
@@ -467,10 +474,10 @@ public class DebuggerTests
 
 		// Check that single stepping works in out-of-line bblocks
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("ss6", (e as StepEvent).Method.Name);
 
@@ -488,12 +495,12 @@ public class DebuggerTests
 		req2.Enable ();
 
 		vm.Resume ();
-		Event e = vm.GetNextEvent ();
+		Event e = GetNextEvent ();
 		Assert.IsTrue (e is MethodEntryEvent);
 		Assert.AreEqual ("ss1", (e as MethodEntryEvent).Method.Name);
 
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is MethodExitEvent);
 		Assert.AreEqual ("ss1", (e as MethodExitEvent).Method.Name);
 
@@ -515,13 +522,13 @@ public class DebuggerTests
 
 		// Enter ss2, ss1 is skipped
 		vm.Resume ();
-		Event e = vm.GetNextEvent ();
+		Event e = GetNextEvent ();
 		Assert.IsTrue (e is MethodEntryEvent);
 		Assert.AreEqual ("ss2", (e as MethodEntryEvent).Method.Name);
 
 		// Breakpoint on ss3, the entry event is no longer reported
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is BreakpointEvent);
 
 		req1.Disable ();
@@ -1086,6 +1093,7 @@ public class DebuggerTests
 	}
 
 	[Test]
+	[Category ("only5")]
 	public void Type_GetValue () {
 		Event e = run_until ("o1");
 		StackFrame frame = e.Thread.GetFrames () [0];
@@ -1118,6 +1126,10 @@ public class DebuggerTests
 		f = t.GetValue (parent.GetField ("base_static_s"));
 		AssertValue ("C", f);
 
+		// thread static field
+		f = t.GetValue (t.GetField ("tls_i"), e.Thread);
+		AssertValue (42, f);
+
 		// Argument checking
 		AssertThrows<ArgumentNullException> (delegate () {
 			t.GetValue (null);
@@ -1367,13 +1379,13 @@ public class DebuggerTests
 		req.Enable ();
 
 		vm.Resume ();
-		var e = vm.GetNextEvent ();
+		var e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("locals2", (e as StepEvent).Method.Name);
 
 		// Execute s = "AB";
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("locals2", (e as StepEvent).Method.Name);
 
@@ -1433,7 +1445,7 @@ public class DebuggerTests
 
 		// invalid stack frames
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("locals2", (e as StepEvent).Method.Name);
 
@@ -1485,7 +1497,7 @@ public class DebuggerTests
 
 		vm.Exit (5);
 
-		var e = vm.GetNextEvent ();
+		var e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (VMDeathEvent), e);
 
 		var p = vm.Process;
@@ -1509,7 +1521,7 @@ public class DebuggerTests
 
 		vm.Dispose ();
 
-		var e = vm.GetNextEvent ();
+		var e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (VMDisconnectEvent), e);
 
 		var p = vm.Process;
@@ -1539,7 +1551,7 @@ public class DebuggerTests
 		
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 
 		l = e.Thread.GetFrames ()[0].Location;
@@ -1550,28 +1562,28 @@ public class DebuggerTests
 		int line_base = l.LineNumber;
 
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		l = e.Thread.GetFrames ()[0].Location;
 		Assert.AreEqual ("ln2", l.Method.Name);
 		Assert.AreEqual (line_base + 6, l.LineNumber);
 
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		l = e.Thread.GetFrames ()[0].Location;
 		Assert.AreEqual ("ln1", l.Method.Name);
 		Assert.AreEqual (line_base + 1, l.LineNumber);
 
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		l = e.Thread.GetFrames ()[0].Location;
 		Assert.AreEqual ("ln3", l.Method.Name);
 		Assert.AreEqual (line_base + 10, l.LineNumber);
 
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		l = e.Thread.GetFrames ()[0].Location;
 		Assert.AreEqual ("ln1", l.Method.Name);
@@ -1624,9 +1636,12 @@ public class DebuggerTests
 	public void AssemblyLoad () {
 		Event e = run_until ("assembly_load");
 
+		var load_req = vm.CreateAssemblyLoadRequest ();
+		load_req.Enable ();
+
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (AssemblyLoadEvent), e);
 		Assert.IsTrue ((e as AssemblyLoadEvent).Assembly.Location.EndsWith ("System.dll"));
 
@@ -1724,6 +1739,16 @@ public class DebuggerTests
 		v = this_obj.InvokeMethod (e.Thread, m, null);
 		AssertValue (42, v);
 
+		// return nullable
+		m = t.GetMethod ("invoke_return_nullable");
+		v = this_obj.InvokeMethod (e.Thread, m, null);
+		Assert.IsInstanceOfType (typeof (StructMirror), v);
+
+		// return nullable null
+		m = t.GetMethod ("invoke_return_nullable_null");
+		v = this_obj.InvokeMethod (e.Thread, m, null);
+		AssertValue (null, v);
+
 		// pass primitive
 		m = t.GetMethod ("invoke_pass_primitive");
 		Value[] args = new Value [] {
@@ -1893,7 +1918,7 @@ public class DebuggerTests
 		StackFrame invoke_frame = null;
 
 		try {
-			e = vm.GetNextEvent ();
+			e = GetNextEvent ();
 			Assert.IsInstanceOfType (typeof (BreakpointEvent), e);
 			// Check stack trace support and invokes
 			var frames = e.Thread.GetFrames ();
@@ -1919,6 +1944,24 @@ public class DebuggerTests
 		o.InvokeMethod (e.Thread, m, null, InvokeOptions.DisableBreakpoints);
 	}
 
+	[Test]
+	public void DisabledExceptionDuringInvoke () {
+		Event e = run_until ("invoke_ex");
+
+		MethodMirror m = entry_point.DeclaringType.GetMethod ("invoke_ex_inner");
+
+		StackFrame frame = e.Thread.GetFrames () [0];
+		var o = frame.GetThis () as ObjectMirror;
+
+		var req = vm.CreateExceptionRequest (null);
+		req.Enable ();
+
+		// Check InvokeOptions.DisableBreakpoints flag
+		o.InvokeMethod (e.Thread, m, null, InvokeOptions.DisableBreakpoints);
+
+		req.Disable ();
+	}
+
 	[Test]
 	public void InvokeSingleThreaded () {
 		vm.Dispose ();
@@ -1953,7 +1996,7 @@ public class DebuggerTests
 				vm.Resume ();
 			}, null);
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.AreEqual (EventType.TypeLoad, e.EventType);
 	}
 
@@ -1970,18 +2013,20 @@ public class DebuggerTests
 
 		Assert.IsTrue (e.Thread.ThreadId > 0);
 
+		Assert.AreEqual (e.Thread.TID, e.Thread.TID);
+
 		vm.EnableEvents (EventType.ThreadStart, EventType.ThreadDeath);
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (ThreadStartEvent), e);
 		var state = e.Thread.ThreadState;
 		Assert.IsTrue (state == ThreadState.Running || state == ThreadState.Unstarted);
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (ThreadDeathEvent), e);
 		Assert.AreEqual (ThreadState.Stopped, e.Thread.ThreadState);
 	}
@@ -2075,7 +2120,7 @@ public class DebuggerTests
 
 		// Step into invoke2
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 		Assert.AreEqual ("invoke2", (e as StepEvent).Method.Name);
 
@@ -2092,7 +2137,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (ExceptionEvent), e);
 		Assert.AreEqual ("OverflowException", (e as ExceptionEvent).Exception.Type.Name);
 
@@ -2108,7 +2153,7 @@ public class DebuggerTests
 		// Skip the throwing of the second OverflowException	   
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (ExceptionEvent), e);
 		Assert.AreEqual ("ArgumentException", (e as ExceptionEvent).Exception.Type.Name);
 		req.Disable ();
@@ -2119,7 +2164,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (ExceptionEvent), e);
 		Assert.AreEqual ("OverflowException", (e as ExceptionEvent).Exception.Type.Name);
 		req.Disable ();
@@ -2130,7 +2175,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (ExceptionEvent), e);
 		Assert.AreEqual ("NullReferenceException", (e as ExceptionEvent).Exception.Type.Name);
 		req.Disable ();
@@ -2141,7 +2186,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (ExceptionEvent), e);
 		Assert.AreEqual ("Exception", (e as ExceptionEvent).Exception.Type.Name);
 		frames = e.Thread.GetFrames ();
@@ -2154,7 +2199,7 @@ public class DebuggerTests
 		sreq.Enable ();
 
 		vm.Resume ();
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (StepEvent), e);
 		frames = e.Thread.GetFrames ();
 		Assert.AreEqual ("exceptions", frames [0].Method.Name);
@@ -2166,6 +2211,85 @@ public class DebuggerTests
 			});
 	}
 
+	[Test]
+	public void ExceptionFilter () {
+		Event e = run_until ("exception_filter");
+
+		MethodMirror m = entry_point.DeclaringType.GetMethod ("exception_filter_filter");
+		Assert.IsNotNull (m);
+
+		vm.SetBreakpoint (m, 0);
+
+		vm.Resume ();
+
+		e = GetNextEvent ();
+		Assert.AreEqual (EventType.Breakpoint, e.EventType);
+		Assert.IsTrue (e is BreakpointEvent);
+		Assert.AreEqual (m.Name, (e as BreakpointEvent).Method.Name);
+
+		var frames = e.Thread.GetFrames ();
+
+		Assert.IsTrue (frames [0].Location.SourceFile.IndexOf ("dtest-app.cs") != -1);
+		Assert.AreEqual ("exception_filter_filter", frames [0].Location.Method.Name);
+
+		Assert.AreEqual (0, frames [1].Location.Method.MetadataToken);
+		Assert.AreEqual (0x0f, frames [1].Location.ILOffset);
+
+		Assert.AreEqual ("exception_filter_method", frames [2].Location.Method.Name);
+		Assert.AreEqual (0x05, frames [2].Location.ILOffset);
+
+		Assert.AreEqual (0, frames [3].Location.Method.MetadataToken, 0);
+		Assert.AreEqual (0, frames [3].Location.ILOffset);
+
+		Assert.AreEqual ("exception_filter", frames [4].Location.Method.Name);
+	}
+
+	[Test]
+	public void ExceptionFilter2 () {
+		vm.Dispose ();
+
+		Start (new string [] { "dtest-excfilter.exe" });
+
+		MethodMirror filter_method = entry_point.DeclaringType.GetMethod ("Filter");
+		Assert.IsNotNull (filter_method);
+
+		MethodMirror test_method = entry_point.DeclaringType.GetMethod ("Test");
+		Assert.IsNotNull (test_method);
+
+		vm.SetBreakpoint (filter_method, 0);
+
+		vm.Resume ();
+
+		var e = GetNextEvent ();
+		Assert.AreEqual (EventType.Breakpoint, e.EventType);
+		Assert.IsTrue (e is BreakpointEvent);
+		Assert.AreEqual (filter_method.Name, (e as BreakpointEvent).Method.Name);
+
+		var frames = e.Thread.GetFrames ();
+
+		Assert.AreEqual (4, frames.Count ());
+
+		Assert.AreEqual (filter_method.Name, frames [0].Location.Method.Name);
+		Assert.AreEqual (20, frames [0].Location.LineNumber);
+		Assert.AreEqual (0, frames [0].Location.ILOffset);
+
+		Assert.AreEqual (test_method.Name, frames [1].Location.Method.Name);
+		Assert.AreEqual (37, frames [1].Location.LineNumber);
+		Assert.AreEqual (0x0b, frames [1].Location.ILOffset);
+
+		Assert.AreEqual (test_method.Name, frames [2].Location.Method.Name);
+		Assert.AreEqual (33, frames [2].Location.LineNumber);
+		Assert.AreEqual (0x05, frames [2].Location.ILOffset);
+
+		Assert.AreEqual (entry_point.Name, frames [3].Location.Method.Name);
+		Assert.AreEqual (14, frames [3].Location.LineNumber);
+		Assert.AreEqual (0x00, frames [3].Location.ILOffset);
+
+		vm.Exit (0);
+
+		vm = null;
+	}
+
 	[Test]
 	public void EventSets () {
 		//
@@ -2211,7 +2335,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (ExceptionEvent), e);
 		Assert.AreEqual ("NullReferenceException", (e as ExceptionEvent).Exception.Type.Name);
 
@@ -2232,7 +2356,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 		
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsInstanceOfType (typeof (AppDomainCreateEvent), e);
 
 		var domain = (e as AppDomainCreateEvent).Domain;
@@ -2244,7 +2368,7 @@ public class DebuggerTests
 
 		while (true) {
 			vm.Resume ();
-			e = vm.GetNextEvent ();
+			e = GetNextEvent ();
 			if (e is BreakpointEvent)
 				break;
 		}
@@ -2267,7 +2391,7 @@ public class DebuggerTests
 
 		while (true) {
 			vm.Resume ();
-			e = vm.GetNextEvent ();
+			e = GetNextEvent ();
 			if (e is BreakpointEvent)
 				break;
 		}
@@ -2280,7 +2404,7 @@ public class DebuggerTests
 		// Run until the unload
 		while (true) {
 			vm.Resume ();
-			e = vm.GetNextEvent ();
+			e = GetNextEvent ();
 			if (e is AssemblyUnloadEvent) {
 				continue;
 			} else {
@@ -2453,7 +2577,7 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is BreakpointEvent);
 
 		var req = vm.CreateStepRequest (e.Thread);
@@ -2463,8 +2587,8 @@ public class DebuggerTests
 
 		vm.Resume ();
 
-		e = vm.GetNextEvent ();
+		e = GetNextEvent ();
 		Assert.IsTrue (e is StepEvent);
 	}
 
-}
\ No newline at end of file
+}
