@@ -351,8 +351,8 @@ namespace MonoTests.System.Threading
 			ev2.WaitOne ();
 
 			Assert.IsFalse (v.TryEnterWriteLock (100));
-			Assert.IsFalse (v.TryEnterReadLock (100));
 			ev.Set ();
+			t1.Join ();
 
 			Assert.IsTrue (v.TryEnterWriteLock (100));
 		}
@@ -412,6 +412,8 @@ namespace MonoTests.System.Threading
 			Assert.IsTrue (v.TryEnterReadLock (100), "#1");
 			Assert.IsTrue (v.TryEnterReadLock (100), "#2");
 			Assert.IsTrue (v.TryEnterReadLock (100), "#3");
+
+			Assert.AreEqual (3, v.RecursiveReadCount);
 		}
 
 		[Test]
@@ -450,6 +452,8 @@ namespace MonoTests.System.Threading
 			Assert.IsTrue (v.TryEnterWriteLock (100), "#1");
 			Assert.IsTrue (v.TryEnterWriteLock (100), "#2");
 			Assert.IsTrue (v.TryEnterWriteLock (100), "#3");
+
+			Assert.AreEqual (3, v.RecursiveWriteCount);
 		}
 
 		[Test]
@@ -482,6 +486,8 @@ namespace MonoTests.System.Threading
 			Assert.IsTrue (v.TryEnterUpgradeableReadLock (100), "#1");
 			Assert.IsTrue (v.TryEnterUpgradeableReadLock (100), "#2");
 			Assert.IsTrue (v.TryEnterUpgradeableReadLock (100), "#3");
+
+			Assert.AreEqual (3, v.RecursiveUpgradeCount);
 		}
 
 		[Test]
