@@ -108,7 +108,7 @@ namespace System.Threading.Tasks
 			this.action              = action == null ? EmptyFunc : action;
 			this.state               = state;
 			this.taskId              = Interlocked.Increment (ref id);
-			this.status              = TaskStatus.Created;
+			this.status              = cancellationToken.IsCancellationRequested ? TaskStatus.Canceled : TaskStatus.Created;
 			this.token               = cancellationToken;
 
 			// Process taskCreationOptions
@@ -142,6 +142,8 @@ namespace System.Threading.Tasks
 		
 		public void Start (TaskScheduler scheduler)
 		{
+			if (status >= TaskStatus.WaitingToRun)
+				throw new InvalidOperationException ("The Task is not in a valid state to be started.");
 			SetupScheduler (scheduler);
 			Schedule ();
 		}
